from ._detector import DetectorBackendPluginBase, DetectorPlotPluginBase, DetectorViewPluginBase
from ._processor import (
    ProcessorBackendPluginBase,
    ProcessorPlotPluginBase,
    ProcessorViewPluginBase,
)
from ._processor_main import processor_main
